/*
** Module   : ContextLocal.java
** Abstract : Container for a context local variable, using the ThreadLocal
**            idiom
**
** Copyright (c) 2006-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20060120   @24009 Created initial version. Contains a context
**                           local variable and exposes access to it using
**                           the ThreadLocal idiom.
** 002 NVS 20060305   @24904 Made modes more reliable and complete. Now
**                           there is STANDALONE mode (no security at all)
**                           and distinct CLIENT and SERVER modes. CLIENT
**                           mode utilizes a static map for storing tokens
**                           so that all client threads share a single
**                           virtual context.
** 003 ECF 20060718   @28091 Fixed memory leak. Storage was not being
**                           cleaned up properly at end of context life
**                           in server and client modes.
** 004 NVS 20070328   @32636 Fixed a bug in CLIENT mode implementation.
**                           The static map for the tokens should be
**                           allocated only once.
** 005 ECF 20070708   @34417 Optimized get() and set(). Only retrieve the
**                           Wrapper object once, rather than checking for
**                           its existence first and then retrieving.
** 006 ECF 20070921   @35244 Reimplemented as generic.
** 007 ECF 20071025   @35525 Made setting of mode threadsafe. Changes
**                           prevent a potential race condition in
**                           setMode().
** 008 ECF 20080506   @38221 Fixed setMode(). Race condition was still
**                           possible, because mode was being set before
**                           all backing support was initialized. Mode is
**                           now set as the last action of setMode().
** 009 CA  20101217          Allow multiple direct session on client nodes - 
**                           the client-side context storage access must be 
**                           synchronized.
** 010 OM  20121016          Added comment description about how to avoid 
**                           issue #1577.
*/

package com.goldencode.p2j.security;

import java.util.*;

/**
 * An implementation of a container for an arbitrary object reference, stored
 * context locally.  Uses the same idiom used by the <code>ThreadLocal</code>
 * class in J2SE, in terms of setting and accessing the reference, and
 * generating an initial value for the container.  This implementation adds
 * an additional feature:  context-specific finalization.  Subclasses can be
 * hooked at the end of a context's life by overriding the {@link #cleanup}
 * method to perform any necessary housekeeping.
 * <p>
 * This class serves a dual role, in that it falls back to a mode which uses
 * a <code>ThreadLocal</code> as backing storage if not running within a
 * security context.  When in server mode, it uses the context-specific
 * token-storing services of <code>SecurityManager</code>.  More recently, a
 * client mode has been added (see below).
 * <p>
 * <b>Server Mode</b><br>
 * This is the mode used in production server use cases (i.e., when a security
 * environment is present).  When running in this mode, the implementation
 * relies upon the following <code>SecurityManager</code> methods:
 * <ul>
 *   <li>{@link SecurityManager#getToken(Object)}
 *   <li>{@link SecurityManager#addToken(Object, Object)}
 *   <li>{@link SecurityManager#removeToken(Object)}
 * </ul>
 * <p>
 * An instance of this class uses itself as the lookup key for these methods.
 * We thus rely upon object identity to guarantee a lookup key is unique.
 * Because of this, this class' {@link #equals(Object)} and {@link
 * #hashCode()} methods are identity-based implementations which may not be
 * overridden.
 * <p>
 * <b>Fallback Mode</b><br>
 * This mode, which stores the object reference in a <code>ThreadLocal</code>,
 * is useful in cases where subsystem code which uses a context local
 * variable is executed in a standalone mode, where no security environment
 * has been created.  In this mode, since there is no security context
 * present, there is no context finalization.  In this mode, the
 * <code>cleanup</code> method will NOT be executed when the local thread
 * expires.  This mode is <i>only</i> used if no <code>SecurityManager</code>
 * instance exists.
 * <p>
 * <b>Client Mode</b></br>
 * This mode is used in a production, client environment (i.e., a security
 * manager is available and it reports that it is a client).  This mode
 * assumes a single context and uses a simple map as backing storage.  As
 * with server mode, an instance of this class uses itself as the lookup key.
 *
 * <p>
 * <strong>Note</strong></br>
 * Do not use <code>ContextLocal</code> object references in the 
 * <code>cleanup()</code> method of </code>Cleanable</code> objects as, at the 
 * time of this call, the context-local variables are already deleted from the 
 * SecurityContext (inside the <code>SecurityContext.cleanup()<code> method).
 * 
 * @param   <T>
 *          Type of object stored context locally.
 */
public class ContextLocal<T>
{
   /** Constant indicating storage mode has not been set */
   private static final int MODE_UNSET = 0;
   
   /** Constant indicating security context based storage mode */
   private static final int MODE_SERVER = 1;
   
   /** Constant indicating client side storage mode */
   private static final int MODE_CLIENT = 2;
   
   /** Constant indicating thread local based storage mode */
   private static final int MODE_STANDALONE = 3;
   
   /** Client side storage for security context tokens */
   private static Map contextStorage = null;
   
   /** Security manager instance (null if no security environment exists) */
   private SecurityManager security = null;
   
   /** An instance of our fallback (ThreadLocal-based) storage mechanism */
   private Fallback<T> fallback = null;
   
   /** Storage mode */
   private volatile int mode = MODE_UNSET;
   
   /**
    * Default constructor.
    */
   public ContextLocal()
   {
   }
   
   /**
    * Get the value of the current context's copy of this context local
    * variable.  The first time this method is invoked, if no value has yet
    * been set into this container, {@link #initialValue} will be invoked.
    *
    * @return  Context local variable.
    */
   public final T get()
   {
      if (mode == MODE_UNSET)
      {
         setMode();
      }
      
      if (mode == MODE_STANDALONE)
      {
         return fallback.get();
      }
      
      T value = null;
      
      if (mode == MODE_SERVER)
      {
         Wrapper<T> w = (Wrapper<T>) security.getToken(this);
         if (w != null)
         {
            value = w.contents;
         }
         else
         {
            value = initialValue();
            set(value);
         }
      }
      else
      {
         synchronized (contextStorage)
         {
            Wrapper<T> w = (Wrapper<T>) contextStorage.get(this);
            if (w != null)
            {
               value = w.contents;
            }
            else
            {
               value = initialValue();
               set(value);
            }
         }
      }
      
      return value;
   }
   
   /**
    * Set the current value of this context local variable.
    *
    * @param   value
    *          Current value.
    */
   public final void set(T value)
   {
      if (mode == MODE_UNSET)
      {
         setMode();
      }
      
      if (mode == MODE_STANDALONE)
      {
         fallback.set(value);
         return;
      }

      if (mode == MODE_SERVER)
      {
         Wrapper<T> w = (Wrapper<T>) security.getToken(this);
         if (w != null)
         {
            w.contents = value;
         }
         else
         {
            security.addToken(this, new Wrapper(value));
         }
      }
      else
      {
         synchronized (contextStorage)
         {
            Wrapper<T> w = (Wrapper<T>) contextStorage.get(this);
            if (w != null)
            {
               w.contents = value;
            }
            else
            {
               contextStorage.put(this, new Wrapper<T>(value));
            }
         }
      }
   }
   
   /**
    * An equality test which relies upon the parent's implementation using
    * object identity.  This method cannot be overridden.
    *
    * @param   o
    *          Object to test for equality against this instance.
    *
    * @return  <code>true</code> if <code>o</code> is the same object
    *          instance as <code>this</code>.
    */
   public final boolean equals(Object o)
   {
      return super.equals(o);
   }
   
   /**
    * Final implementation of <code>hashCode</code>, which delegates to the
    * superclass and cannot be overridden.
    *
    * @return  Hash code for this object.
    */
   public final int hashCode()
   {
      return super.hashCode();
   }
   
   /**
    * This method will be invoked to retrieve the initial value of this
    * context local variable, the first time {@link #get} is invoked (but
    * only if {@link #set} was not previously invoked.  The returned value
    * will be stored as the variable's initial value.
    * <p>
    * This default implementation simply returns <code>null</code>.
    * Subclasses which require a different initial value must override this
    * method.
    *
    * @return  <code>null</code> by default.
    */
   protected T initialValue()
   {
      return null;
   }
   
   /**
    * This method is invoked at the end of the security context's life.  It
    * should be overridden by subclasses which need to perform cleanup at
    * this point.  Implementors should not use security specific services at
    * this point.
    *
    * @param   value
    *          The last value this context local variable contained.
    */
   protected void cleanup(T value)
   {
   }
   
   /**
    * Set the mode used to store the local copy of the variable.  If no
    * security manager has been created, thread local storage is used, else
    * security context storage is used.
    */
   private synchronized void setMode()
   {
      // Resolve possible race condition.
      if (mode != MODE_UNSET)
      {
         return;
      }
      
      security = SecurityManager.getInstance();

      if (security == null)
      {
         fallback = new Fallback<T>();
         mode = MODE_STANDALONE;
         return;
      }

      if (security.isServer())
      {
         mode = MODE_SERVER;
      }
      else
      {
         if (contextStorage == null)
         {
            contextStorage = new HashMap();
         }
         mode = MODE_CLIENT;
      }
   }
   
   /**
    * An instance of this class is used to wrap the variable value when it is
    * stored in the security context.  Its primary purpose is to implement
    * the {@link #cleanup} method, which the current security context will
    * invoke at the end of its life.  Only one instance of this class is
    * created per context, per variable, the first time a value needs to be
    * stored.  Thereafter, each time the variable's value is updated, the
    * contents of this wrapper are merely reset.
    */
   private class Wrapper<E>
   implements Cleanable
   {
      /** Data value stored by this context local variable */
      private E contents = null;
      
      /**
       * Convenience constructor which assigns contents.
       *
       * @param   contents
       *          Data value stored by this context local variable
       */
      Wrapper(E contents)
      {
         this.contents = contents;
      }
      
      /**
       * Invoked at security context end of life.  Removes this token from
       * storage and invokes the enclosing class' <code>cleanup</code>
       * method, to give subclasses a chance to perform housekeeping.
       */
      public void cleanup()
      {
         ContextLocal.this.cleanup((T) contents);
         switch (mode)
         {
            case MODE_SERVER:
               security.removeToken(ContextLocal.this);
               break;
            case MODE_CLIENT:
               contextStorage.remove(ContextLocal.this);
               break;
         }
      }
   }
   
   /**
    * An implementation of <code>ThreadLocal</code> which is used as the
    * backing storage for this context local variable when no security
    * environment is present.  This implementation is necessary to allow
    * the fallback thread local variable's <code>initialValue</code>
    * method to invoke the enclosing class' implementation (which may have
    * been overridden by a subclass).
    */
   private class Fallback<E>
   extends ThreadLocal<E>
   {
      /**
       * Reroute an invocation of thread local's <code>initialValue</code>
       * implementation to the enclosing class' method of the same name.
       *
       * @return  The initial value returned by the enclosing class'
       *          <code>initialValue</code> implementation.
       */
      protected E initialValue()
      {
         return (E) ContextLocal.this.initialValue();
      }
   }
}
